#!/usr/bin/python
#
# Copyright (c) 2015-2017, Virtuozzo International GmbH
#
# This file is based on the same file from the Virtuozzo Core project.
# You can redistribute it file and/or modify it under the terms
# of the GNU General Public License as published by the Free Software
# Foundation; either version 2 of the License, or (at your option) any
# later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.
#
# Our contact details: Virtuozzo International GmbH, Vordergasse 59, 8200
# Schaffhausen, Switzerland.
#

from ConfigParser import SafeConfigParser
import glob
import os
from StringIO import StringIO
from subprocess import Popen, PIPE, call, check_output
import re


CONFIG_PATH = "/etc/sysconfig/network-scripts"
SECTION_NAME = "section"
BACKUP_PREFIX = "vz_preserved-"

class FakeSection(object):
    """Wrapper for input stream inserting sections."""
    def __init__(self, f):
        self.f = f
        self.section_read = False

    def readline(self):
        if self.section_read:
            return self.f.readline()
        self.section_read = True
        return "[%s]\n" % SECTION_NAME


class NoSectionConfigParser(SafeConfigParser):
    """Config parser for files without section headers.

    Has dict-like interface, but supports ConfigParser-like as well.
    """
    def __init__(self):
        SafeConfigParser.__init__(self)
        self.optionxform = str
        self.add_section(SECTION_NAME)

    def _read(self, fp, fpname):
        SafeConfigParser._read(self, FakeSection(fp), fpname)

    def get(self, option, default):
        try:
            return SafeConfigParser.get(self, SECTION_NAME, option)
        except:
            return default

    def set(self, option, value):
        SafeConfigParser.set(self, SECTION_NAME, option, value)

    def __setitem__(self, option, value):
        self.set(option, value)

    def __getitem__(self, option):
        return SafeConfigParser.get(self, SECTION_NAME, option)

    def delete(self, option):
        if not self.remove_option(SECTION_NAME, option):
            raise ValueError("No such option")

    def items(self):
        return SafeConfigParser.items(self, SECTION_NAME)

    def write(self, fp):
        for (key, value) in self._sections[SECTION_NAME].items():
            if key == "__name__":
                continue
            if value or (self._optcre == self.OPTCRE):
                key = "=".join((key, str(value)))
            fp.write("%s\n" % (key))

    def __iter__(self):
        return (item[0] for item in self.items())


class NetworkServiceManager(object):
    """Controls network services.

    Remembers status that was on creation.
    Does NOT update status according to commands executed.
    Works with a single 'active' network service.
    """

    # Known network services in decreasing priority.
    known_services = ["NetworkManager", "network"]

    def __init__(self):
        self.services = {}
        self.active_service = None
        for service in self.known_services:
            status = self.get_network_service_status(service)
            self.services[service] = status
            if status["running"]:
                self.active_service = service

    @staticmethod
    def get_network_service_status(netservice):
        p = Popen(["systemctl", "show", netservice,
                   "--property=LoadState,ActiveState,SubState"], stdout=PIPE, stderr=PIPE)
        stdout, stderr = p.communicate()
        if p.returncode != 0:
            print "Systemctl exited with code %d" % p.returncode
            return {"loaded": False, "running": False}
        cp = NoSectionConfigParser()
        cp.readfp(StringIO(stdout))
        try:
            if cp["LoadState"] != "loaded":
                return {"loaded": False, "running": False}
            if cp["ActiveState"] == "active" and cp["SubState"] == "running":
                return {"loaded": True, "running": True}
            return {"loaded": True, "running": False}
        except:
            return {"loaded": False, "running": False}

    @staticmethod
    def call_network_service(netservice, command):
        """Execute command via systemctl."""
        if netservice not in NetworkServiceManager.known_services:
            raise ValueError("Network service unknown")
        return call(["systemctl", command, netservice])

    def stop_networking(self):
        if not self.active_service:
            # Silently ignore
            return 0
        return self.call_network_service(self.active_service, "stop")

    def start_networking(self):
        if not self.active_service:
            # Silently ignore
            return 0
        return self.call_network_service(self.active_service, "start")


def dequote(s):
    """Remove outer quotes."""
    s = s.strip()
    if len(s) >= 2 and s[0] == s[-1] and s[0] in ["\'", "\""]:
        s = s[1:-1]
    return s

def read_configs():
    interfaces = {}
    for ifcfg in glob.glob("ifcfg-*"):
        cp = NoSectionConfigParser()
        cp.read(ifcfg)
        interfaces[ifcfg.replace("ifcfg-", "")] = cp
    return interfaces

def write_configs(to_write):
    written = []
    error = False
    for cp, filename, backup in to_write:
        try:
            if backup:
                os.rename(filename, backup)
                written.append((filename, backup))
                cp.write(open(filename, "w"))
            else:
                cp.write(open(filename, "w"))
                written.append((filename, backup))
        except:
            error = True
            break

    if error:
        # Fallback.Restore original configs.
        for filename, backup in written:
            # Best effort.
            try:
                if backup:
                    os.rename(backup, filename)
                else:
                    os.remove(filename)
            except:
                pass


'''
Fix 'MASTER' parameter - we want it to use device names, not UUIDs
'''
def rename_masters():
    # Read interface configs.
    os.chdir(CONFIG_PATH)
    interfaces = read_configs()
    # device <-> uuid mapping
    dev_uuid_map = {}

    # First, map uuids to devices
    for iface, cp in interfaces.items():
        uuid = dequote(cp.get("UUID", ""))
        if not uuid:
            continue
        dev = cp.get("DEVICE", "")
        if dev:
            dev_uuid_map[uuid] = dev

    # Now go throug configs and fix the ones that specify master using UUID
    to_write = []
    for iface, cp in interfaces.items():
        master = dequote(cp.get("MASTER", ""))
        if master in dev_uuid_map:
            cp.set("MASTER", dev_uuid_map[master])
        to_write.append((cp, "ifcfg-" + iface, "/var/log/anaconda/vz-renamed-ifcfg" + iface))

    if to_write and not os.path.exists("/var/log/anaconda"):
        os.mkdir("/var/log/anaconda")

    return to_write


def fix_devices():
    netmanager = NetworkServiceManager()

    to_write = rename_masters()
    if to_write:
        netmanager.stop_networking()
        write_configs(to_write)
        netmanager.start_networking()


if __name__ == "__main__":
    fix_devices()

